ACTIVITY_RENAME = {
    'redcap_repeat_instance': 'ID',
    'activity_description': 'DESCRIPTION',
    'activity_datetime': 'DATETIME',
    'activity_event': 'EVENT',
    'activity_field': 'FIELD',
    'activity_result': 'RESULT',
    'activity_scan': 'SCAN',
    'activity_subject': 'SUBJECT',
    'activity_session': 'SESSION',
    'activity_type': 'CATEGORY',
}

ANALYSES_RENAME = {
    'redcap_repeat_instance': 'ID',
    'analysis_name': 'NAME',
    'analysis_include': 'SUBJECTS',
    'analysis_processor': 'PROCESSOR',
    'analysis_input': 'INPUT',
    'analysis_output': 'OUTPUT',
    'analyses_complete': 'COMPLETE',
    'analysis_status': 'STATUS',
    'analysis_covars': 'COVARS',
    'analysis_notes': 'NOTES',
}

ISSUES_RENAME = {
    'redcap_repeat_instance': 'ID',
    'issue_date': 'DATETIME',
    'issue_description': 'DESCRIPTION',
    'issue_event': 'EVENT',
    'issue_field': 'FIELD',
    'issue_scan': 'SCAN',
    'issue_session': 'SESSION',
    'issue_subject': 'SUBJECT',
    'issue_type': 'CATEGORY',
}

PROCESSING_RENAME = {
    'redcap_repeat_instance': 'ID',
    'processor_file': 'FILE',
    'processor_filter': 'FILTER',
    'processor_args': 'ARGS',
}

TASKS_RENAME = {
    'task_assessor': 'ASSESSOR',
    'task_status': 'STATUS',
    'task_inputlist': 'INPUTLIST',
    'task_var2val': 'VAR2VAL',
    'task_memreq': 'MEMREQ',
    'task_walltime': 'WALLTIME',
    'task_procdate': 'PROCDATE',
    'task_timeused': 'TIMEUSED',
    'task_memused': 'MEMUSED',
    'task_yamlfile': 'YAMLFILE',
    'task_userinputs': 'USERINPUTS',
    'task_failcount': 'FAILCOUNT',
    'task_yamlupload': 'YAMLUPLOAD',
}

REPORTS_RENAME = {
    'progress_name': 'NAME',
    'progress_datetime': 'DATE',
    'progress_pdf': 'PDF',
    'progress_zip': 'DATA',
    'double_resultspdf': 'PDF',
    'double_resultsfile': 'DATA',
    'double_name': ' NAME',
    'double_datetime': 'DATE',
}

DISABLE_STATTYPES = [
    'fmri_rest_v2',
    'fmri_roi_v1',
    'struct_preproc_noflair_v1',
    'fmri_nback_v2',
    'Multi_Atlas_v3', 
    'fmri_roi_v2',
]

COLUMNS = {
    'activity': [
        'PROJECT', 'SUBJECT', 'SESSION', 'SCAN', 'ID', 'DESCRIPTION',
        'DATETIME', 'EVENT', 'FIELD', 'CATEGORY', 'RESULT', 'STATUS', 'SOURCE'],
    'assessors': [
        'PROJECT', 'SUBJECT', 'SESSION', 'SESSTYPE', 'NOTE', 'DATE', 'SITE',
        'ASSR', 'PROCSTATUS', 'PROCTYPE', 'JOBDATE', 'QCSTATUS',
        'QCDATE', 'QCBY', 'XSITYPE', 'INPUTS', 'MODALITY', 'full_path'],
    'issues': [
        'ID', 'DATETIME', 'PROJECT', 'CATEGORY',
        'SUBJECT', 'SESSION', 'SCAN ', 'DESCRIPTION',
        'EVENT', 'FIELD', 'STATUS'],
    'scans': [
        'PROJECT', 'SUBJECT', 'SESSION', 'SESSTYPE', 'TRACER', 'NOTE', 'DATE', 'SITE',
        'SCANID', 'SCANTYPE', 'QUALITY', 'RESOURCES', 'MODALITY', 'XSITYPE', 'full_path'],
    'processing': [
        'ID', 'PROJECT', 'TYPE', 'FILTER', 'FILE', 'CUSTOM', 'ARGS', 'YAMLUPLOAD', 'EDIT'],
    'subjects': [
        'PROJECT', 'SUBJECT', 'AGE', 'SEX', 'RACE'],
    'tasks': [
        'ID', 'PROJECT', 'STATUS', 'PROCTYPE', 'MEMREQ', 'WALLTIME',
        'TIMEUSED', 'MEMUSED', 'ASSESSOR', 'PROCDATE', 'INPUTLIST', 'VAR2VAL',
        'IMAGEDIR', 'JOBTEMPLATE', 'YAMLFILE', 'YAMLUPLOAD', 'USERINPUTS', 'FAILCOUNT'],
    'analyses': ['PROJECT', 'ID', 'NAME', 'STATUS', 'COMPLETE', 'EDIT', 'INPUT', 'OUTPUT', 'NOTES', 'SUBJECTS', 'PROCESSOR'],
    'processors': ['ID', 'PROJECT', 'TYPE', 'EDIT', 'FILE', 'FILTER', 'ARGS'],
    'sgp': ['PROJECT', 'SUBJECT', 'ASSR', 'PROCSTATUS', 'PROCTYPE', 'QCSTATUS', 'INPUTS', 'DATE', 'XSITYPE'],
    'reports': ['TYPE', 'PROJECT', 'NAME', 'DATE', 'PDF', 'DATA'],
}


PROCLIB = {
    'AMYVIDQA_v2': {
        'short_descrip': 'Regional Amyloid SUVR using cerebellum as reference.',
        'inputs_descrip': 'T1w MRI processed with FreeSurfer (FS7_v1), Amyloid PET',
        'procurl': 'https://github.com/ccmvumc/AMYVIDQA',
        'stats_subset': ['compositegm_suvr', 'cblmtot_suvr', 'cblmwm_suvr', 'cblmgm_suvr', 'hippocampus_suvr']
    },
    'BFC_v2': {
        'short_descrip': 'Basal Forebrain Volumes.',
        'inputs_descrip': 'T1w MRI',
        'procurl': 'https://github.com/ccmvumc/BFC',
        'stats_subset': ['CH4_L_VOL', 'CH4_R_VOL']
    },
    'BrainAgeGap_v2': {
        'short_descrip': 'Predicted age of brain.',
        'inputs_descrip': 'T1w MRI parcellated with BrainColor atlas',
        'procurl': 'https://pubmed.ncbi.nlm.nih.gov/32948749/',
    },
    'FALLYPRIDEQA_v1':{
        'short_descrip': 'Fallypride QA with Regional SUVR using whole cerebellum as reference.',
        'inputs_descrip': 'T1w MRI processed with FreeSurfer (FS7_v1), Fallypride PET',
        'procurl': 'https://github.com/bud42/FALLYPRIDEQA',
        'stats_subset': ['antcing_suvr', 'compositegm_suvr', 'cblmgm_suvr', 'cblmwm_suvr', 'cblmtot_suvr'],
    },
    'fmri_bct_v1': {
        'short_descrip': 'Brain Connectivity Toolbox measures.',
        'inputs_descrip': 'Resting MRI processed with fmri_roi_v2',
        'procurl': 'https://github.com/REMBRANDT-study/fmri_bct',
        'stats_subset': ['Shen268_thr0p3_degree', 'Schaefer400_thr0p3_degree'],
    },
    'fmri_msit_v2': {
        'short_descrip': 'fMRI MSIT task pre-processing and 1st-Level analysis.',
        'inputs_descrip': 'T1w MRI, MSIT fMRI, E-prime EDAT',
        'procurl': 'https://github.com/REMBRANDT-study/fmri_msit',
        'stats_subset': ['con_amyg_mean', 'inc_amyg_mean', 'med_pct_outliers', 'con_bnst_mean', 'inc_bnst_mean'],
    },
    'fmri_rest_v2': {
        'short_descrip': 'fMRI Resting State pre-processing.',
        'inputs_descrip': 'T1w MRI, Resting State fMRI',
        'procurl': 'https://github.com/REMBRANDT-study/fmri_rest',
    },
    'fmri_roi_v2': {
        'short_descrip': 'Regional measures of functional connectivity',
        'inputs_descrip': 'Resting State fMRI processed with fmri_rest_v2',
        'procurl': 'https://github.com/REMBRANDT-study/fmri_roi',
    },
    'FS7sclimbic_v0': {
        'short_descrip': 'FreeSurfer 7 ScLimbic - volumes of subcortical limbic regions including Basal Forebrain.',
        'inputs_descrip': 'T1w MRI processed with FreeSurfer (FS7_v1)',
        'procurl': 'https://surfer.nmr.mgh.harvard.edu/fswiki/ScLimbic',
        'stats_subset': ['Left-Basal-Forebrain', 'Right-Basal-Forebrain', 'Left-Nucleus-Accumbens', 'Right-Nucleus-Accumbens', 'eTIV'],
    },
    'FEOBVQA_v2': {
        'short_descrip': 'Regional SUVR using Supra-ventricular White Matter as reference.',
        'inputs_descrip': 'T1w MRI processed with FreeSurfer (FS7_v1), FEOBV PET',
        'procurl': 'https://github.com/ccmvumc/FEOBVQA',
        'stats_subset': ['cblmwm_suvr', 'compositegm_suvr', 'cblmgm_suvr'],
    },
    'FS7_v1': {
        'short_descrip': 'FreeSurfer 7 recon-all - whole brain parcellation, surfaces, cortical thickness.',
        'inputs_descrip': 'T1w MRI',
        'procurl': 'https://github.com/bud42/FS7',
    },
    'FS7HPCAMG_v1': {
        'short_descrip': 'FreeSurfer 7 hippocampus & amygdala sub-region volumes.',
        'inputs_descrip': 'T1w processed with FreeSurfer (FS7_v1)',
        'procurl': 'https://github.com/bud42/FS7HPCAMG_v1',
        'stats_subset': ['hpchead_lh', 'hpchead_rh', 'hpcbody_lh', 'hpcbody_rh', 'hpctail_lh', 'hpctail_rh'],
    },
    'LST_v1': {
        'short_descrip': 'Lesion Segmentation Toolbox - white matter lesion volumes.',
        'inputs_descrip': 'T1w MRI, FLAIR MRI',
        'procurl': 'https://github.com/ccmvumc/LST1',
    },
    'SAMSEG_v1': {
        'short_descrip': 'Runs SAMSEG from FreeSurfer 7.2 to get White Matter Lesion Volume.',
        'inputs_descrip': 'T1w MRI processed with FreeSurfer (FS7_v1), FLAIR MRI',
        'procurl': 'https://surfer.nmr.mgh.harvard.edu/fswiki/Samseg',
    },
    'fmriqa_v4': {
        'short_descrip': 'Functional MRI QA',
        'stats_subset': ['dvars_mean', 'fd_mean'],
    },
    'fmri_emostroop_v2': {
        'short_descrip': 'fMRI EmoStroop Pre-processing and 1st-Level',
        'inputs_descrip': 'T1w MRI, fMRI, EDAT',
        'procurl': 'https://github.com/ccmvumc/fmri_emostroop:v2',
        'stats_subset': ['lhSFG2_incgtcon', 'rhSFG2_incgtcon', 'overall_rt_mean'],
    },
}

STATLIB = {
    'FALLYPRIDEQA_v1': {
        'accumbens_suvr': 'Accumbens regional mean SUVR normalized by Whole Cerebellum',
        'amygdala_suvr': 'Amygdala regional mean SUVR normalized by Whole Cerebellum',
        'antcing_suvr': 'Anterior Cingulate regional mean SUVR normalized by Whole Cerebellum',
        'antflobe_suvr': 'Anterior Frontal Lobe mean SUVR normalized by Whole Cerebellum',
        'caudate_suvr': 'Caudate regional mean SUVR normalized by Whole Cerebellum',
        'cblmgm_suvr': 'Cerebellum Gray Matter regional mean SUVR normalized by Whole Cerebellum',
        'cblmtot_suvr': 'Cerebellum Total regional mean SUVR normalized by Whole Cerebellum',
        'cblmwm_suvr': 'Cerebellum White Matter regional mean SUVR normalized by Whole Cerebellum',
        'compositegm_suvr': 'Composite Gray Matter regional mean SUVR normalized by Whole Cerebellum',
        'cortwm_suvr': 'Cortical White Matter regional mean SUVR normalized by Whole Cerebellum',
        'hippocampus_suvr': 'Hippocampus regional mean SUVR normalized by Whole Cerebellum',
        'latplobe_suvr': 'Lateral Parietal Lobe regional mean SUVR normalized by Whole Cerebellum',
        'lattlobe_suvr': 'Lateral Temporal Lobe regional mean SUVR normalized by Whole Cerebellum',
        'mofc_suvr': 'Medial Orbito-frontal Cortex regional mean SUVR normalized by Whole Cerebellum',
        'pallidum_suvr': 'Pallidum regional mean SUVR normalized by Whole Cerebellum',
        'postcing_suvr': 'Posterior Cingulate regional mean SUVR normalized by Whole Cerebellum',
        'putamen_suvr': 'Putamen regional mean SUVR normalized by Whole Cerebellum',
        'thalamus_suvr': 'Thalamus regional mean SUVR normalized by Whole Cerebellum',
        'ventraldc_suvr': 'Ventral Diencephalon regional mean SUVR normalized by Whole Cerebellum',
    },
    'FS7_v1': {
        'fs7_etiv': 'Estimated Total Intracranial Volume',
        'fs7_hpc_lh': 'Hippocampus Left Hemisphere Volume',
        'fs7_hpc_rh': 'Hippocampus Right Hemisphere Volume',
        'fs7_latvent_lh': 'Lateral Ventricle Left Hemisphere Volume',
        'fs7_latvent_rh': 'Lateral Ventricle Right Hemisphere Volume',
        'fs7_stnv': 'Supra-tentorial not ventricles Volume',
        'fs7_supflobe_lh': 'Superior Frontal Lobe Left Hemisphere Thickness',
        'fs7_supflobe_rh': 'Superior Frontal Lobe Right Hemisphere Thickness',
    },
    'FEOBVQA_v2': {
        'antcing_suvr': 'Anterior Cingulate SUVR normalized by Supra-ventricular White Matter',
        'antflobe_suvr': 'Anterior Frontal Lobe SUVR normalized by Supra-ventricular White Matter',
        'cblmgm_suvr': 'Cerebellar Gray Matter SUVR normalized by Supra-ventricular White Matter',
        'cblmwm_suvr': 'Cerebellar White Matter SUVR normalized by Supra-ventricular White Matter',
        'compositegm_suvr': 'Composite Gray Matter SUVR normalized by Supra-ventricular White Matter',
        'cblmgm_suvr': 'Cerebellar Gray Matter SUVR normalized by Supra-ventricular White Matter',
        'cortwm_eroded_suvr': 'Eroded Cortical White Matter SUVR normalized by Supra-ventricular White Matter',
        'latplobe_suvr': 'Lateral Parietal Lobe SUVR normalized by Supra-ventricular White Matter',
        'lattlobe_suvr': 'Lateral Temporal Lobe SUVR normalized by Supra-ventricular White Matter',
        'postcing_suvr': 'Posterior Cingulate SUVR normalized by Supra-ventricular White Matter',
    },
    'SAMSEG_v1': {
        'samseg_lesions': 'whole brain White Matter Lesion Volume in cubic millimeters',
        'samseg_sbtiv': 'segmentation-based (estimated) Total Intracranial Volume in cubic millimeters',
    },
    'FS7HPCAMG_v1': {
        'hpcbody_lh': 'Hippocampus Body Left Hemisphere Volume in cubic millimeters',
        'hpcbody_rh': 'Hippocampus Body Right Hemisphere Volume in cubic millimeters',
        'hpchead_lh': 'Hippocampus Head Left Hemisphere Volume in cubic millimeters',
        'hpchead_rh': 'Hippocampus Head Right Hemisphere Volume in cubic millimeters',
        'hpctail_lh': 'Hippocampus Tail Left Hemisphere Volume in cubic millimeters',
        'hpctail_rh': 'Hippocampus Tail Right Hemisphere Volume in cubic millimeters',
    },
    'LST_v1': {
        'wml_volume': 'White Matter Lesion Volume',
    },
    'fmriqa_v4': {
        'dvars_mean': 'DVARS, framewise signal',
        'fd_mean': 'Framewise Displacement',
    },
    'struct_preproc_v1': {
        'Volume1': 'Gray Matter',
        'Volume2': 'White Matter',
        'Volume3': 'CSF',
    },
    'fmri_bct_v1': {
        'Schaefer400_thr0p1_deg': 'Degree'
    }
}



import logging
import os
import pandas as pd

from ...garjus import Garjus


logger = logging.getLogger('dashboard.processors.data')


def get_filename():
    datadir = f'{Garjus().cachedir()}/DATA'
    filename = f'{datadir}/processorsdata.pkl'

    try:
        os.makedirs(datadir)
    except FileExistsError:
        pass

    return filename


def run_refresh(filename, projects):
    df = get_data(projects)

    save_data(df, filename)

    return df


def load_options():
    garjus = Garjus()
    proj_options = garjus.projects()

    return proj_options


def load_data(projects, refresh=False):
    filename = get_filename()

    if refresh or not os.path.exists(filename):
        run_refresh(filename, projects)

    logger.info('reading data from file:{}'.format(filename))
    return read_data(filename)


def read_data(filename):
    df = pd.read_pickle(filename)
    return df


def save_data(df, filename):
    # save to cache
    df.to_pickle(filename)


def get_data(projects):
    df = pd.DataFrame()
    garjus = Garjus()

    pid = garjus.redcap_pid()

    # Load
    df = garjus.processing_protocols(projects)

    df['FILE'] = df['FILE'].apply(os.path.basename)

    df = df.sort_values(['PROJECT', 'FILE'])

    # Make edit link
    df['EDIT'] = 'https://redcap.vanderbilt.edu/redcap_v13.9.3/DataEntry/index.php?pid=' + \
        str(pid) + \
        '&page=processing&id=' + \
        df['PROJECT'] + \
        '&instance=' + \
        df['ID'].astype(str)

    return df


def filter_data(df):
    # TBD

    return df



"""Analyses."""
import logging
import os, shutil
import tempfile
import subprocess as sb

import pandas as pd


logger = logging.getLogger('garjus.analyses')


def _download_zip(xnat, uri, zipfile):
    # Build the uri to download
    _uri = uri + '?format=zip&structure=simplified'

    response = xnat.get(_uri, stream=True)
    with open(zipfile, 'wb') as f:
        shutil.copyfileobj(response.raw, f)

    return zipfile


def _download_file_stream(xnat, uri, dst):

    response = xnat.get(uri, stream=True)

    with open(dst, 'wb') as f:
        shutil.copyfileobj(response.raw, f)

    return dst


def update(garjus, projects=None):
    """Update analyses."""
    for p in (projects or garjus.projects()):
        if p in projects:
            logger.debug(f'updating analyses:{p}')
            _update_project(garjus, p)


def _update_project(garjus, project):
    analyses = garjus.analyses([project], download=True)

    if len(analyses) == 0:
        logger.debug(f'no open analyses for project:{project}')
        return

    # Handle each record
    for i, a in analyses.iterrows():
        aname = a['NAME']

        if not a.get('PROCESSOR', False):
            logger.debug(f'no processor:{aname}')
            continue

        if a['COMPLETE'] != '2':
            logger.debug(f'skipping complete not set:{aname}')
            continue

        if a['STATUS'] == 'READY':
            logger.debug(f'skipping done:{aname}')
            continue

        logger.info(f'updating analysis:{aname}')
        _update(garjus, a)


def update_analysis(
    garjus,
    project,
    analysis_id
):
    """Update analysis."""

    # Run it
    _update(garjus, analysis)


def _has_outputs(garjus, analysis):
    project = analysis['PROJECT']
    analysis_id = analysis['ID']
    resource = f'{project}_{analysis_id}'
    res_uri = f'/projects/{project}/resources/{resource}'
    output_zip = f'{project}_{analysis_id}_OUTPUTS.zip'

    res = garjus.xnat().select(res_uri)

    file_list = res.files()
    if output_zip in file_list:
        logger.info(f'found:{output_zip}')
        return True
    else:
        return False


def _has_inputs(garjus, analysis):
    project = analysis['PROJECT']
    analysis_id = analysis['ID']
    resource = f'{project}_{analysis_id}'
    res_uri = f'/projects/{project}/resources/{resource}'
    inputs_zip = f'{project}_{analysis_id}_INPUTS.zip'

    res = garjus.xnat().select(res_uri)

    file_list = res.files()
    if inputs_zip in file_list:
        logger.info(f'found:{inputs_zip}')
        return True
    else:
        logger.info(f'inputs not found:{res_uri}:{inputs_zip}')

        return False


def _update(garjus, analysis):
    with tempfile.TemporaryDirectory() as tempdir:
        inputs_dir = f'{tempdir}/INPUTS'
        outputs_dir = f'{tempdir}/OUTPUTS'

        if _has_outputs(garjus, analysis):
            logger.debug(f'outputs exist')
        elif _has_inputs(garjus, analysis):
            logger.debug(f'inputs exist')
        else:
            _make_dirs(inputs_dir)
            _make_dirs(outputs_dir)

            # Create new inputs
            logger.info(f'downloading analysis inputs to {inputs_dir}')
            _download_inputs(garjus, analysis, inputs_dir)

            logger.info(f'uploading analysis inputs zip')
            try:
                dst = upload_inputs(
                    garjus,
                    analysis['PROJECT'], 
                    analysis['ID'],
                    tempdir)

                logger.debug(f'set analysis inputs')
                garjus.set_analysis_inputs(
                    analysis['PROJECT'],
                    analysis['ID'],
                    dst)
            except Exception as err:
                logger.error(f'upload failed')
                return

            logger.info(f'running analysis')
            _run(garjus, analysis, tempdir)

            # Set STATUS
            logger.info(f'set analysis status')
            garjus.set_analysis_status(
                analysis['PROJECT'],
                analysis['ID'],
                'READY')

    # That is all
    logger.info(f'analysis done!')


def _run(garjus, analysis, tempdir):
    processor = analysis['PROCESSOR']

    # Run commmand and upload output
    command = processor.get('command', None)
    if command is None:
        logger.debug('no command found')
        return

    # Run steps
    logger.info('running analysis steps...')

    # Get the container name or path
    container = processor['command']['container']
    for c in processor['containers']:
        if c['name'] == container:
            container = c['source']

        if 'path' in c:
            print(c['path'], 'not yet')
            continue

    if shutil.which('singularity'):
        command_mode = 'singularity'
    elif shutil.which('docker'):
        command_mode = 'docker'
    else:
        logger.error('command mode not found, cannot run container command')
        return

    logger.debug(f'command mode is {command_mode}')

    if command_mode == 'singularity':
        # Build the command string
        cmd = f'singularity run -c -e -B {tempdir}/INPUTS:/INPUTS -B {tempdir}/OUTPUTS:/OUTPUTS {container}'
    elif command_mode == 'docker':
        if container.startswith('docker://'):
            # Remove docker prefix
            container = container.split('docker://')[1]

        # Build the command string
        cmd = f'docker run -it --rm -v {tempdir}/INPUTS:/INPUTS -v {tempdir}/OUTPUTS:/OUTPUTS {container}'

    # Run it
    logger.info(cmd)
    os.system(cmd)

    # Upload it
    logger.info(f'uploading output')
    dst = upload_outputs(garjus, analysis['PROJECT'], analysis['ID'], tempdir)
    garjus.set_analysis_outputs(analysis['PROJECT'], analysis['ID'], dst)


def run_analysis(garjus, project, analysis_id):
    analysis = garjus.load_analysis(project, analysis_id)

    _run(garjus, analysis)

    # That is all
    logger.info(f'analysis done!')


def upload_outputs(garjus, project, analysis_id, tempdir):
    # Upload output_zip Project Resource on XNAT named with
    # the project and analysis id as PROJECT_ID, e.g. REMBRANDT_1
    resource = f'{project}_{analysis_id}'
    res_uri = f'/projects/{project}/resources/{resource}'
    outputs_zip = f'{tempdir}/{project}_{analysis_id}_OUTPUTS.zip'

    # Zip output
    logger.info(f'zipping output to {outputs_zip}')
    sb.run(['zip', '-r', outputs_zip, 'OUTPUTS'], cwd=tempdir)

    logger.debug(f'connecting to xnat resource:{res_uri}')
    res = garjus.xnat().select(res_uri)

    logger.debug(f'uploading file to xnat resource:{outputs_zip}')
    res.file(os.path.basename(outputs_zip)).put(
        outputs_zip,
        overwrite=True,
        params={"event_reason": "analysis upload"})

    uri = f'{garjus.xnat_host()}/data{res_uri}/files/{os.path.basename(outputs_zip)}'

    return uri


def upload_inputs(garjus, project, analysis_id, tempdir):
    # Upload to Project Resource on XNAT named with
    # the project and analysis id as PROJECT_ID, e.g. REMBRANDT_1
    resource = f'{project}_{analysis_id}'
    res_uri = f'/projects/{project}/resources/{resource}'
    inputs_zip = f'{tempdir}/{project}_{analysis_id}_INPUTS.zip'

    logger.info(f'zipping inputs {tempdir} to {inputs_zip}')
    sb.run(['zip', '-r', inputs_zip, 'INPUTS'], cwd=tempdir)

    assert(os.path.isfile(inputs_zip))

    logger.debug(f'connecting to xnat resource:{res_uri}')
    res = garjus.xnat().select(res_uri)

    logger.debug(f'uploading file to xnat resource:{inputs_zip}')
    res.file(os.path.basename(inputs_zip)).put(
        inputs_zip,
        overwrite=True,
        params={"event_reason": "analysis upload"})

    uri = f'{garjus.xnat_host()}/data{res_uri}/files/{os.path.basename(inputs_zip)}'

    return uri


def _sessions_from_scans(scans):
    return scans[[
        'PROJECT',
        'SUBJECT',
        'SESSION',
        'SESSTYPE',
        'DATE',
        'SITE'
    ]].drop_duplicates()


def _sessions_from_assessors(assessors):
    return assessors[[
        'PROJECT',
        'SUBJECT',
        'SESSION',
        'SESSTYPE',
        'DATE',
        'SITE'
    ]].drop_duplicates()


def _download_file(garjus, proj, subj, sess, assr, res, fmatch, dst):
    # Make the folders for this file path
    _make_dirs(os.path.dirname(dst))

    # Connect to the resource on xnat
    r = garjus.xnat().select_assessor_resource(proj, subj, sess, assr, res)

    # TODO: apply regex or wildcards in fmatch
    # res_obj.files()[0].label()).get(fpath)
    # res.files().label()

    r.file(fmatch).get(dst)

    return dst


def _download_sgp_resource_zip(xnat, project, subject, assessor, resource, outdir):
    reszip = '{}_{}.zip'.format(assessor, resource)
    respath = 'data/projects/{}/subjects/{}/experiments/{}/resources/{}/files'
    respath = respath.format(project, subject, assessor, resource)

    # Download the resource as a zip file
    download_zip(respath, reszip)

    # Unzip the file to output dir
    logger.debug(f'unzip file {rezip} to {outdir}')
    with zipfile.ZipFile(reszip) as z:
        z.extractall(outdir)

    # TODO: check downloaded files, compare/size/md5 to xnat

    # Delete the zip
    os.remove(reszip)


def _download_sgp_file(garjus, proj, subj, assr, res, fmatch, dst):
    # Make the folders for this file path
    _make_dirs(os.path.dirname(dst))

    # Download the file
    uri = f'data/projects/{proj}/subjects/{subj}/experiments/{assr}/resources/{res}/files/{fmatch}'
    _download_file_stream(garjus.xnat(), uri, dst)


def _download_resource(garjus, proj, subj, sess, assr, res, dst):
    # Make the folders for destination path
    _make_dirs(dst)

    # Connect to the resource on xnat
    r = garjus.xnat().select_assessor_resource(proj, subj, sess, assr, res)

    # Download resource and extract
    r.get(dst, extract=True)

    return dst


def _download_sgp_resource(garjus, proj, subj, assr, res, dst):
    # Make the folders for destination path
    _make_dirs(dst)

    # Connect to the resource on xnat
    r = garjus.xnat().select_sgp_assessor(proj, subj, assr).resource(res)

    # Download extracted
    r.get(dst, extract=True)

    return dst


def _make_dirs(dirname):
    try:
        os.makedirs(dirname)
    except FileExistsError:
        pass


def _download_subject_assessors(garjus, subj_dir, sgp_spec, proj, subj, sgp):

    sgp = sgp[sgp.SUBJECT == subj]

    for k, a in sgp.iterrows():

        assr = a.ASSR

        for assr_spec in sgp_spec:
            logger.debug(f'assr_spec={assr_spec}')

            assr_types = assr_spec['types'].split(',')

            logger.debug(f'assr_types={assr_types}')

            if a.PROCTYPE not in assr_types:
                logger.debug(f'skip assr, no match on type={assr}:{a.PROCTYPE}')
                continue

            for res_spec in assr_spec['resources']:

                try:
                    res = res_spec['resource']
                except (KeyError, ValueError) as err:
                    logger.error(f'reading resource:{err}')
                    continue

                if 'fmatch' in res_spec:
                    # Download files
                    for fmatch in res_spec['fmatch'].split(','):
                        # Where shall we save it?
                        dst = f'{subj_dir}/{assr}/{res}/{fmatch}'

                        # Have we already downloaded it?
                        if os.path.exists(dst):
                            logger.debug(f'exists:{dst}')
                            continue

                        # Download it
                        logger.info(f'download file:{proj}:{subj}:{assr}:{res}:{fmatch}')
                        try:
                            _download_sgp_file(
                                garjus,
                                proj,
                                subj,
                                assr,
                                res,
                                fmatch,
                                dst
                            )
                        except Exception as err:
                            logger.error(f'{subj}:{assr}:{res}:{fmatch}:{err}')
                            import traceback
                            traceback.print_exc()
                            raise err
                else:
                    # Download whole resource

                    # Where shall we save it?
                    dst = f'{subj_dir}/{assr}'

                    # Have we already downloaded it?
                    if os.path.exists(os.path.join(dst, res)):
                        logger.debug(f'exists:{dst}')
                        continue

                    # Download it
                    logger.info(f'download resource:{proj}:{subj}:{assr}:{res}')
                    try:
                        _download_sgp_resource(
                            garjus,
                            proj,
                            subj,
                            assr,
                            res,
                            dst
                        )
                    except Exception as err:
                        logger.error(f'{subj}:{assr}:{res}:{err}')
                        raise err


def _download_subject(garjus, subj_dir, subj_spec, proj, subj, sessions, assessors, sgp):

    #  subject-level assessors
    sgp_spec = subj_spec.get('assessors', None)
    if sgp_spec:
        logger.debug(f'download_sgp={subj_dir}')
        _download_subject_assessors(garjus, subj_dir, sgp_spec, proj, subj, sgp)

    # Download the subjects sessions
    for sess_spec in subj_spec.get('sessions', []):
        sess_types = sess_spec['types'].split(',')

        for i, s in sessions[sessions.SUBJECT == subj].iterrows():
            sess = s.SESSION

            # Apply session type filter
            if s.SESSTYPE not in sess_types:
                logger.debug(f'skip session, no match on type={sess}:{s.SESSTYPE}')
                continue

            sess_dir = f'{subj_dir}/{sess}'
            logger.debug(f'download_session={sess_dir}')
            _download_session(
                garjus, sess_dir, sess_spec, proj, subj, sess, assessors)


def _download_session(garjus, sess_dir, sess_spec, proj, subj, sess, assessors):
    # get the assessors for this session
    sess_assessors = assessors[assessors.SESSION == sess]

    for k, a in sess_assessors.iterrows():
        assr = a.ASSR

        for assr_spec in sess_spec.get('assessors', []):
            logger.debug(f'assr_spec={assr_spec}')

            assr_types = assr_spec['types'].split(',')

            logger.debug(f'assr_types={assr_types}')

            if a.PROCTYPE not in assr_types:
                logger.debug(f'skip assr, no match on type={assr}:{a.PROCTYPE}')
                continue

            for res_spec in assr_spec['resources']:

                try:
                    res = res_spec['resource']
                except (KeyError, ValueError) as err:
                    logger.error(f'reading resource:{err}')
                    continue

                if 'fmatch' in res_spec:
                    # Download files
                    for fmatch in res_spec['fmatch'].split(','):

                        # Where shall we save it?
                        dst = f'{sess_dir}/{assr}/{res}/{fmatch}'

                        # Have we already downloaded it?
                        if os.path.exists(dst):
                            logger.debug(f'exists:{dst}')
                            continue

                        # Download it
                        logger.info(f'download file:{proj}:{subj}:{sess}:{assr}:{res}:{fmatch}')
                        try:
                            _download_file(
                                garjus,
                                proj,
                                subj,
                                sess,
                                assr,
                                res,
                                fmatch,
                                dst
                            )
                        except Exception as err:
                            logger.error(f'{subj}:{sess}:{assr}:{res}:{fmatch}:{err}')
                            raise err
                else:
                    # Download whole resource

                    # Where shall we save it?
                    dst = f'{sess_dir}/{assr}'

                    # Have we already downloaded it?
                    if os.path.exists(os.path.join(dst, res)):
                        logger.debug(f'exists:{dst}')
                        continue

                    # Download it
                    logger.info(f'download resource:{proj}:{subj}:{sess}:{assr}:{res}')
                    try:
                        _download_resource(
                            garjus,
                            proj,
                            subj,
                            sess,
                            assr,
                            res,
                            dst
                        )
                    except Exception as err:
                        logger.error(f'{subj}:{sess}:{assr}:{res}:{err}')
                        raise err


def download_analysis_inputs(garjus, project, analysis_id, download_dir):

    logger.debug(f'download_analysis_inputs:{project}:{analysis_id}:{download_dir}')

    analysis = garjus.load_analysis(project, analysis_id)

    _download_inputs(garjus, analysis, download_dir)


def _download_inputs(garjus, analysis, download_dir):
    errors = []

    project = analysis['PROJECT']

    logger.info('loading project data')
    assessors = garjus.assessors(projects=[project])
    scans = garjus.scans(projects=[project])
    sgp = garjus.subject_assessors(projects=[project])

    sessions = pd.concat([
        _sessions_from_scans(scans),
        _sessions_from_assessors(assessors)
    ])
    sessions = sessions.drop_duplicates()

    # Which subjects to include?
    subjects = analysis['SUBJECTS'].splitlines()

    logger.debug(f'subjects={subjects}')

    # What to download for each subject?
    subj_spec = analysis['PROCESSOR']['inputs']['xnat']['subjects']

    logger.debug(f'subject spec={subj_spec}')

    for subj in subjects:
        logger.debug(f'subject={subj}')

        # Make the Subject download folder
        subj_dir = f'{download_dir}/{subj}'
        _make_dirs(subj_dir)

        # Download the subject as specified in subj_spec
        try:
            logger.debug(f'_download_subject={subj}')
            _download_subject(
                garjus,
                subj_dir,
                subj_spec,
                project,
                subj,
                sessions,
                assessors,
                sgp)
        except Exception as err:
            logger.debug(err)
            errors.append(subj)
            continue

    # report what's missing
    if errors:
        logger.info(f'errors{errors}')
    else:
        logger.info(f'download complete with no errors!')

    logger.debug('done!')



from setuptools import setup, find_packages

setup(
    name="garjus",
    version="1.1.3",
    author="Brian D. Boyd",
    author_email="brian.d.boyd@vumc.org",
    description="Python package for managing imaging data in REDCap and XNAT",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/ccmvumc/garjus",
    packages=find_packages(),
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
    ],
    install_requires=[
        "pandas",
        "pycap",
        "pyxnat",
        "dax",
        "click",
        "sphinx",
        "pydot",
        "plotly",
        "dash_bootstrap_components",
        "dash_bootstrap_templates",
    ],
    entry_points={"console_scripts": ["garjus = garjus.cli:cli"]},
)



"""

update will create any missing

"""
import tempfile
import csv
import logging
import glob
import os

import numpy as np
import pandas as pd


logger = logging.getLogger(__name__)


def update(garjus, projects):
    """Update project progress."""
    for p in projects:
        proctypes = garjus.stattypes(p)

        if not proctypes:
            logger.debug(f'no proctypes for stats project:{p}')
            continue

        logger.debug(f'stats updating project:{p},proctypes={proctypes}')
        update_project(garjus, p, proctypes)

        if 'BrainAgeGap_v2' in proctypes:
            logger.debug('getting bag_age_gap')
            _get_bag(garjus, p)


def update_project(garjus, project, proctypes):
    """Update stats for project proctypes."""

    logger.debug(f'loading existing stats:{project}')
    try:
        # Get list of assessors already in stats
        existing = garjus.stats_assessors(project, proctypes)
    except Exception:
        logger.error(f'failed to load existing stats, check key:{project}')
        return

    # Get list of all assessors
    logger.debug(f'loading existing assessors:{project}')

    dfa = garjus.assessors([project], proctypes)
    logger.debug(f'total assessors:{len(dfa)}')

    # Filter to remove already uploaded
    dfa = dfa[~dfa['ASSR'].isin(existing)]
    logger.debug(f'assessors after filtering out already uploaded:{len(dfa)}')

    # Filter to only COMPLETE
    dfa = dfa[dfa['PROCSTATUS'] == 'COMPLETE']
    logger.debug(f'assessors after filtering only COMPLETE:{len(dfa)}')

    # Filter to not Failed
    dfa = dfa[dfa['QCSTATUS'] != 'Failed']
    logger.debug(f'assessors after filtering out QC Failed:{len(dfa)}')

    # Iterate xnat assessors
    for r in dfa.sort_values('ASSR').to_dict('records'):
        try:
            update_assessor(
                garjus,
                r['PROJECT'],
                r['SUBJECT'],
                r['SESSION'],
                r['ASSR'],
            )
        except ConnectionError as err:
            logger.info(err)
            logger.info('waiting a minute')
            os.sleep(60)


def update_assessor(garjus, proj, subj, sess, assr):
    """Update assessor stats."""
    logger.debug(f'uploading assessor stats:{assr}')
    with tempfile.TemporaryDirectory() as tmpdir:
        try:
            _dir = garjus.get_source_stats(proj, subj, sess, assr, tmpdir)
        except Exception as err:
            logger.warn(f'could not get stats:{assr}:{err}')
            return

        _stats = transform_stats(_dir)
        garjus.set_stats(proj, subj, sess, assr, _stats)


def transform_stats(stats_dir):
    """Transform stats from directory of files to dict."""
    data = {}

    if os.path.exists(f'{stats_dir}/stats.csv'):
        data = _load_stats(f'{stats_dir}/stats.csv')
    elif os.path.exists(f'{stats_dir}/stats.txt'):
        data = _load_stats(f'{stats_dir}/stats.txt')
    elif os.path.exists(f'{stats_dir}/fmriqa_stats.csv'):
        data.update(_load_stats_tall(f'{stats_dir}/fmriqa_stats.csv'))
    elif len(glob.glob(f'{stats_dir}/*.txt')) > 0:
        for txt_path in glob.iglob(f'{stats_dir}/*.txt'):
            data.update(_load_stats_tall(txt_path))
    else:
        # Handle proctypes that output multiple csv
        for csv_path in glob.iglob(f'{stats_dir}/*.csv'):
            data.update(_load_stats_wide(csv_path))

    return data


def _isfloat(num):
    try:
        float(num)
        return True
    except ValueError:
        return False


def _load_stats_wide(filename):
    data = {}

    with open(filename, newline='') as f:
        # Connect csv reader
        reader = csv.reader(f)

        # Load header from first line
        header = next(reader)

        # Read data from subsequent lines
        for line in reader:
            for i, v in enumerate(line):
                data[header[i]] = v

    return data


def _load_stats_tall(filename):
    data = {}
    rows = []

    try:
        with open(filename) as f:
            rows = f.readlines()

        for r in rows:
            (k, v) = r.strip().replace('=', ',').split(',')
            data[k] = v
    except ValueError:
        logger.error(f'cannot load stats file')
        return {}

    return data


def _load_stats(filename):
    lines = []
    with open(filename) as f:
        lines = f.readlines()

    if ('=' in lines[0]) or (len(lines) >= 3 and len(lines[0].split(',')) <= 3):
        return _load_stats_tall(filename)
    else:
        return _load_stats_wide(filename)


def _get_bag(garjus, project):

    # Get subjects with DOB
    subjects = garjus.subjects(project, include_dob=True)
    if 'DOB' not in subjects.columns:
        logger.debug('DOB not found, cannot calculate bag_age_gap')
        return

    # Get BAG stats
    stats = garjus.stats(project, proctypes=['BrainAgeGap_v2'])

    if stats.empty:
        return

    # Merge in DOB
    stats = pd.merge(
        stats, subjects[['DOB']], left_on='SUBJECT', right_index=True)

    if 'bag_age_gap' in stats:
        # Only rows without existing bag_age_gap
        stats = stats[stats.bag_age_gap.isna()]

    # Calculate age at scan
    stats['SCANDAYS'] = pd.to_datetime(stats['DATE']) - stats['DOB']
    stats['BAGDAYS'] = (stats['bag_age_pred'].astype(float) * 365.25).astype('timedelta64[D]')
    stats['bag_age_gap'] = (stats['BAGDAYS'] - stats['SCANDAYS'])/np.timedelta64(1, 'Y')

    # Batch upload new stats
    for i, s in stats.iterrows():
        logger.debug('set bag_age_gap', s.ASSR)
        garjus.set_stats(
            project,
            s.SUBJECT,
            s.SESSION,
            s.ASSR,
            {'bag_age_gap': s.bag_age_gap})



import logging
import tempfile
import io

import pandas as pd
import numpy as np

from ...utils_redcap import field2events, download_file


def parse_tar(filename):
    encoding = 'utf-8'
    skiprows = 0
    first_field = 'ExperimentName'

    # Determine how many rows to skip prior to header
    with io.open(filename, encoding=encoding) as f:
        for line in f:
            if line.startswith(first_field):
                break
            else:
                skiprows += 1

    # Load Data
    df = pd.read_table(
        filename, sep='\t', encoding=encoding, skiprows=skiprows, header=0)

    return df


def load_tarfile(filename):
    df = parse_tar(filename)

    # Remove whitespace that we know is there
    df['EmoWord'] = df['EmoWord'].str.strip()

    # Separate good and bad words
    dfb = df[df['EmoWord'] == 'bad']
    dfg = df[df['EmoWord'] == 'good']

    # Calculate summary data
    data = {}
    data['tar_hit_good'] = len(dfg[(dfg['RecallListB.CRESP'] == 'm') & (dfg['RecallListB.RESP'] == 'm')])
    data['tar_miss_good'] = len(dfg[(dfg['RecallListB.CRESP'] == 'm') & (dfg['RecallListB.RESP'] == 'x')])
    data['tar_fa_good'] = len(dfg[(dfg['RecallListB.CRESP'] == 'x') & (dfg['RecallListB.RESP'] == 'm')])
    data['tar_cr_good'] = len(dfg[(dfg['RecallListB.CRESP'] == 'x') & (dfg['RecallListB.RESP'] == 'x')])
    data['tar_hit_bad'] = len(dfb[(dfb['RecallListB.CRESP'] == 'm') & (dfb['RecallListB.RESP'] == 'm')])
    data['tar_miss_bad'] = len(dfb[(dfb['RecallListB.CRESP'] == 'm') & (dfb['RecallListB.RESP'] == 'x')])
    data['tar_fa_bad'] = len(dfb[(dfb['RecallListB.CRESP'] == 'x') & (dfb['RecallListB.RESP'] == 'm')])
    data['tar_cr_bad'] = len(dfb[(dfb['RecallListB.CRESP'] == 'x') & (dfb['RecallListB.RESP'] == 'x')])

    return data


def etl_tar(project, record_id, event_id):
    data = None
    tab_file = None
    tab_field = 'tat_recall_conv'

    with tempfile.TemporaryDirectory() as tmpdir:
        tab_file = f'{tmpdir}/{record_id}-{event_id}-{tab_field}.txt'

        # Download converted tab-delimited file
        try:
            download_file(
                project,
                record_id,
                tab_field,
                tab_file,
                event_id=event_id)
        except Exception as err:
            logging.error(f'download failed:{record_id}:{event_id}:{err}')
            return False

        # Transform data
        try:
            data = _transform(tab_file)
            data[project.def_field] = record_id
            data['redcap_event_name'] = event_id
        except Exception as err:
            logging.error(f'transform failed:{record_id}:{event_id}:{err}')
            return False

    # Load the data back to redcap
    try:
        _response = project.import_records([data])
        assert 'count' in _response
        logging.info(f'TAR uploaded:{record_id}')
    except AssertionError as err:
        logging.error(f'TAR upload:{record_id}:{err}')
        return False

    return True


def _transform(filename):
    data = {}

    # Load the data
    logging.info(f'extracting tar:{filename}')
    tar_data = load_tarfile(filename)
    if tar_data is None:
        logging.error('extract failed')
    else:
        # Transform the data and force strings for redcap
        data['recall_hit_good'] = str(tar_data['tar_hit_good'])
        data['recall_miss_good'] = str(tar_data['tar_miss_good'])
        data['recall_fa_good'] = str(tar_data['tar_fa_good'])
        data['recall_cr_good'] = str(tar_data['tar_cr_good'])
        data['recall_hit_bad'] = str(tar_data['tar_hit_bad'])
        data['recall_miss_bad'] = str(tar_data['tar_miss_bad'])
        data['recall_fa_bad'] = str(tar_data['tar_fa_bad'])
        data['recall_cr_bad'] = str(tar_data['tar_cr_bad'])

    return data


def process_project(project):
    results = []
    tab_field = 'tat_recall_conv'
    done_field = 'recall_hit_good'
    def_field = project.def_field
    fields = [def_field, tab_field, done_field]
    id2subj = {}
    events = field2events(project, tab_field)
    sec_field = project.export_project_info()['secondary_unique_field']
    if sec_field:
        rec = project.export_records(fields=[def_field, sec_field])
        id2subj = {x[def_field]: x[sec_field] for x in rec if x[sec_field]}
    else:
        rec = project.export_records(fields=[def_field])
        id2subj = {x[def_field]: x[def_field] for x in rec if x[def_field]}

    # Get records
    rec = project.export_records(fields=fields, events=events)

    # Process each record
    for r in rec:
        record_id = r[def_field]
        event_id = r['redcap_event_name']
        subj = id2subj.get(record_id)

        # Check for converted file
        if not r[tab_field]:
            logging.debug(f'not yet converted:{record_id}:{subj}:{event_id}')
            continue

        # Determine if ETL has already been run
        if r[done_field]:
            logging.debug(f'already ETL:{record_id}:{subj}:{event_id}')
            continue

        # Do the ETL
        result = etl_tar(project, record_id, event_id)
        if result:
            results.append({
                'result': 'COMPLETE',
                'category': 'etl_traitadjrecall',
                'description': 'etl_traitadjrecall',
                'subject': subj,
                'event': event_id,
                'field': tab_field})

    return results



"""Shared settings for dashboard used across apps."""

# These are used to set colors of graphs
RGB_DKBLUE = 'rgb(59,89,152)'
RGB_BLUE = 'rgb(66,133,244)'
RGB_GREEN = 'rgb(15,157,88)'
RGB_YELLOW = 'rgb(244,160,0)'
RGB_RED = 'rgb(219,68,55)'
RGB_PURPLE = 'rgb(160,106,255)'
RGB_GREY = 'rgb(200,200,200)'
RGB_PINK = 'rgb(255,182,193)'
RGB_LIME = 'rgb(17, 180, 101)'

# These are used to set color of html tables via style argument
HEX_LBLUE = '#DAEBFF'
HEX_LGREE = '#DCFFDA'
HEX_LYELL = '#FFE4B3'
HEX_LREDD = '#FFDADA'
HEX_LGREY = '#EBEBEB'
HEX_LPURP = '#D1C0E5'
HEX_LPINK = '#FFB6C1'

# Give each status a color to display
QASTATUS2COLOR = {
    'PASS': RGB_GREEN,
    'NQA': RGB_LIME,
    'NPUT': RGB_YELLOW,
    'FAIL': RGB_RED,
    'NONE': RGB_GREY,
    'JOBF': RGB_PINK,
    'JOBR': RGB_BLUE}

DEFAULT_COLOR = 'rgba(0,0,0,0.5)'

LINE_COLOR = 'rgba(50,50,50,0.9)'

STATUS2RGB = dict(zip(
    ['WAITING', 'PENDING', 'RUNNING', 'COMPLETE', 'FAILED', 'UNKNOWN', 'JOBF', 'UPLOADING', 'FAIL', 'PASS', 'NQA', 'NPUT'],
    [RGB_GREY, RGB_YELLOW, RGB_GREEN, RGB_BLUE, RGB_RED, RGB_PURPLE, RGB_PINK, RGB_LIME, RGB_RED, RGB_GREEN, RGB_LIME, RGB_YELLOW]))

STATUS2HEX = dict(zip(
    ['WAITING', 'PENDING', 'RUNNING', 'COMPLETE', 'FAILED', 'UNKNOWN', 'JOBF'],
    [HEX_LGREY, HEX_LYELL, HEX_LGREE, HEX_LBLUE, HEX_LREDD, HEX_LPURP, HEX_LPINK]))

# These are used to make progress reports
ASTATUS2COLOR = {
    'PASS': RGB_GREEN,
    'NPUT': RGB_YELLOW,
    'FAIL': RGB_RED,
    'NQA': RGB_LIME,
    'NONE': RGB_GREY,
    'COMPLETE': RGB_BLUE,
    'UNKNOWN': RGB_PURPLE}

GWIDTH = 1000



"""Compare 1st and 2nd Entry."""
import os
import logging
from datetime import datetime

import pandas as pd
from fpdf import FPDF


# Compares two REDCap projects, as First and Second where Second is the
# subset of fields that should be double-entered. Outputs a PDF file
# and an excel file. The PDF provides an overview/summary of the Excel
# file. The excel file contains the specific missing and conflicting items.
# Output files are named with the specified prefix.


# Names and descriptions of the sheets in the excel output file
SHEETS = [{
'name': 'Mismatches',
'description': 'Mismatches is the list of discrepancies between the First \
and Second REDCap projects with one row per mismatch.'''
},
{
'name': 'MissingSubjects',
'description': 'MissingSubjects is the list of subjects that are found in \
the First REDCap project but are completely missing from the Second.'
},
{
'name': 'MissingEvents',
'description': 'MissingEvents is the list of subject events that are found \
in the First REDCap project, but are completely missing from the Second.'
},
{
'name': 'MissingValues',
'description': 'MissingEvents is the list of values that are found \
in the First REDCap project, but are blank in the Second.'
},
{
'name': 'FieldsCompare',
'description': 'FieldsCompare is the list of fields that are INCLUDED for \
comparison. This list includes all fields in FieldsCommon excluding those \
in Fields2ndNan.'
},
{
'name': 'FieldsCommon',
'description': 'FieldsCommon is the list of fields that are found in both \
First and Second REDCap projects.'
},
{
'name': 'Fields1stOnly',
'description': 'Fields1stOnly is the list of fields that are found only in \
the First REDCap. These fields are EXCLUDED from comparisons.'
},
{
'name': 'Fields2ndOnly',
'description': 'Fields2ndOnly is the list of fields that are found only in \
the Second REDCap project. This should be an empty list.'
},
{
'name': 'Fields2ndNan',
'description': 'Fields2ndNan is the list of fields that are found in both \
REDCap projects, but all values are blank in the Second REDCap. This list \
should be empty.'
}]


# Our custom PDF file format
class MYPDF(FPDF):
    def footer(self):
        today = datetime.now().strftime("%Y-%m-%d")
        self.date = today
        self.title = 'Double Entry'
        self.subtitle = '{}'.format(datetime.now().strftime("%B %Y"))

        self.set_y(-0.35)
        self.set_x(0.5)

        # Write date, title, page number
        self.set_font('helvetica', size=10)
        self.set_text_color(100, 100, 100)
        self.set_draw_color(100, 100, 100)
        self.line(x1=0.2, y1=10.55, x2=8.3, y2=10.55)
        self.cell(w=1, txt=self.date)
        self.cell(w=5, align='C', txt=self.title)
        self.cell(w=2.5, align='C', txt=f'{self.page_no()} of {{nb}}')


def make_pdf(results, filename):
    logging.debug('making PDF')

    # Initialize a new PDF letter size and shaped
    pdf = MYPDF(orientation="P", unit='in', format='letter')
    pdf.add_page()

    # Give it a title at the top
    title = 'Double Data Entry Comparison'
    pdf.set_font('helvetica', size=18)
    pdf.cell(w=7, h=0.5, txt=title, border=0, ln=1)

    # Iterate the heading section
    for key, val in results['session'].items():
        pdf.set_font('helvetica', size=13)
        pdf.cell(w=1.3, h=.6, txt=key, border=0)
        pdf.set_font('courier', size=14)
        pdf.cell(w=6, h=.6, txt=val, border=1, ln=1)

    # Show results counts section
    counts = results['counts']

    pdf.ln(0.1)
    pdf.set_font('helvetica', size=14)
    pdf.cell(1, 0.4, 'RESULTS:', ln=1)

    txt = 'Matches'
    val = str(counts['matches'])
    pdf.set_font('courier', size=12)
    pdf.cell(w=2, h=.3, txt=txt, border=0)
    pdf.cell(w=1, h=.3, txt=val, border=1, align='C', ln=1)

    txt = 'Mismatches'
    val = str(counts['mismatches'])
    dsc = 'see Mismatches sheet'
    pdf.set_font('courier', size=12)
    pdf.cell(w=2, h=.3, txt=txt, border=0)
    pdf.cell(w=1, h=.3, txt=val, border=1, align='C')
    pdf.set_font('courier', size=9)
    pdf.cell(w=5, h=.3, txt=dsc, border=0, ln=1)

    txt = 'Missing Subjects'
    val = str(counts['missing_subjects'])
    dsc = 'see MissingSubjects sheet'
    pdf.set_font('courier', size=12)
    pdf.cell(w=2, h=.3, txt=txt, border=0)
    pdf.cell(w=1, h=.3, txt=val, border=1, align='C')
    pdf.set_font('courier', size=9)
    pdf.cell(w=5, h=.3, txt=dsc, border=0, ln=1)

    txt = 'Missing Events'
    val = str(counts['missing_events'])
    dsc = 'see MissingEvents sheet'
    pdf.set_font('courier', size=12)
    pdf.cell(w=2, h=.3, txt=txt, border=0)
    pdf.cell(w=1, h=.3, txt=val, border=1, align='C')
    pdf.set_font('courier', size=9)
    pdf.cell(w=5, h=.3, txt=dsc, border=0, ln=1)

    txt = 'Missing Values'
    val = str(counts['missing_values'])
    dsc = 'see MissingValues sheet'
    pdf.set_font('courier', size=12)
    pdf.cell(w=2, h=.3, txt=txt, border=0)
    pdf.cell(w=1, h=.3, txt=val, border=1, align='C')
    pdf.set_font('courier', size=9)
    pdf.cell(w=5, h=.3, txt=dsc, border=0, ln=1)

    pdf.ln(0.25)

    pdf.set_font('helvetica', size=10)
    _txt = 'The sheets in the excel file are:'
    pdf.cell(w=7.5, h=0.3, txt=_txt, border='T', ln=1)

    # Add sheet descriptions
    for s in SHEETS:
        add_sheet_description(pdf, s['name'], s['description'])

    # Save to file
    logging.info(f'saving PDF to file:{filename}')
    try:
        pdf.output(filename)
    except Exception as err:
        logging.error(f'error while saving PDF:{filename}:{err}')


def add_sheet_description(pdf, name, description):
    # Write the name and description to the PDF
    pdf.set_font(style='B', size=8)
    pdf.cell(w=1.2, h=0.4, txt=name, border=0)
    pdf.set_font(style='')
    pdf.multi_cell(w=6.3, h=0.4, txt=description, border='B', ln=1, align='L')


def get_fields(p1, p2):
    # Get all the records so we can check for all nan
    df = p2.export_records(format_type='df')

    common_fields = sorted(list(set(p1.field_names) & set(p2.field_names)))
    p1_only_fields = sorted(list(set(p1.field_names) - set(p2.field_names)))
    p2_only_fields = sorted(list(set(p2.field_names) - set(p1.field_names)))
    p2_used_fields = sorted(list((df.dropna(how='all', axis=1)).columns))
    p2_nan_fields = sorted(list(set(common_fields) - set(p2_used_fields)))
    compare_fields = sorted(list(set(common_fields) - set(p2_nan_fields)))

    ignore_fields = ['subj_num', 'ddes_12child', 'entry_consider_list',
    'entryconsid_c_list' 'pe_date', 'physical_specif', 'physical_yesno',
    'pvas_date', 'pvas_interfere', 'pvas_pain']

    compare_fields = [x for x in compare_fields if x not in ignore_fields]

    fields = {
        'compare': compare_fields,
        'common': common_fields,
        'p1_only': p1_only_fields,
        'p2_only': p2_only_fields,
        'p2_nan': p2_nan_fields,
    }

    return fields


def write_sheet(data, writer, name):
    # Format as string to avoid formatting mess
    df = pd.DataFrame(data=data, dtype=str)

    # Sort by all columns in order
    df = df.sort_values(by=list(df.columns))

    # Write the dataframe to the excel file
    df.to_excel(writer, sheet_name=name, index=False)

    # Auto-adjust columns' width
    col_fmt = writer.book.add_format({'num_format': '@'})
    for column in df:
        col_width = max(df[column].astype(str).map(len).max(), len(column))
        col_idx = df.columns.get_loc(column)
        writer.sheets[name].set_column(col_idx, col_idx, col_width, col_fmt)

    # Ignore errors
    _range = '{}{}:{}{}'.format('A', 1, 'F', len(df) + 1)
    writer.sheets[name].ignore_errors({'number_stored_as_text': _range})


def write_excel(info, outfile):
    # Write each sheet to excel
    with pd.ExcelWriter(outfile) as w:
        if info['mismatches']:
            write_sheet(info['mismatches'], w, 'Mismatches')

        if info['missing_subjects']:
            write_sheet(info['missing_subjects'], w, 'MissingSubjects')

        if info['missing_events']:
            write_sheet(info['missing_events'], w, 'MissingEvents')

        if info['missing_values']:
            write_sheet(info['missing_values'], w, 'MissingValues')

        write_sheet(info['fields']['common'], w, 'FieldsCommon')
        write_sheet(info['fields']['compare'], w, 'FieldsCompare')

        write_sheet(info['fields']['p1_only'], w, 'Fields1stOnly')
        write_sheet(info['fields']['p2_only'], w, 'Fields2ndOnly')
        write_sheet(info['fields']['p2_nan'], w, 'Fields2ndNan')


def _boring_record(record, compare_fields):
    for f in compare_fields:
        if record.get(f, False):
            # There's a value
            return False

    return True


def export_all_records(project, step=100):
    records = []

    # Get a list of unique IDs
    _def = project.def_field
    ids = [r[_def] for r in project.export_records(fields=[_def])]
    ids = sorted(list(set(ids)))

    # How many?
    count = len(ids)

    # Load in chunks of step size
    for i in range(0, count, step):
        r = project.export_records(records=ids[i:i + step])
        records.extend(r)

    return records


def compare_projects(p1, p2):
    # Compares two redcap projects and returns the results
    results = {}
    missing_subjects = []
    missing_events = []
    missing_values = []
    mismatches = []
    sec_field = None
    match_count = 0

    # Create index of record ID to subject ID in p1
    def_field = p1.def_field
    sec_field = p1.export_project_info()['secondary_unique_field']
    def_field2 = p2.def_field
    sec_field2 = p2.export_project_info()['secondary_unique_field']

    if sec_field:
        rec = p1.export_records(fields=[def_field, sec_field])
        id2subj1 = {x[def_field]: x[sec_field] for x in rec if x[sec_field]}

        # Create index of subject ID to record ID in p2
        rec = p2.export_records(fields=[def_field2, sec_field2])
        subj2id2 = {x[sec_field2]: x[def_field2] for x in rec if x[sec_field2]}
    else:
        rec = p1.export_records(fields=[def_field])
        id2subj1 = {x[def_field]: x[def_field] for x in rec if x[def_field]}
        rec = p2.export_records(fields=[def_field2])
        subj2id2 = {x[def_field2]: x[def_field2] for x in rec if x[def_field2]}

    # Determine which fields to compare
    fields = get_fields(p1, p2)
    compare_fields = fields['compare']
    date_fields = [x for x in compare_fields if ('date' in x or '_dt' in x)]
    logging.debug(f'date_fields={date_fields}')

    # Get the records from the First project
    records1 = export_all_records(p1)

    # Set subject id
    for i, r in enumerate(records1):
        try:
            rid1 = r[def_field]
            records1[i]['sid'] = id2subj1[rid1]
        except KeyError as err:
            logging.debug(f'blank subject ID for record:{rid1}:{err}')
            records1[i]['sid'] = rid1

    # Sort by subject
    records1 = sorted(records1, key=lambda x: x['sid'])

    # Compare each record
    for r1 in records1:
        sid = r1['sid']
        eid = r1['redcap_event_name']
        name1 = str(r1.get('redcap_repeat_instrument', ''))
        num1 = str(r1.get('redcap_repeat_instance', ''))
        r2 = None
        name2 = ''
        num2 = ''

        if _boring_record(r1, compare_fields):
            # nothing to compare
            continue

        # Check that we already found as missing
        if sid in missing_subjects:
            # Skip this subject, already missing
            logging.info(f'subject already missing:{sid}')
            continue

        if (sid, eid) in missing_events:
            # Skip this event, already missing
            logging.info(f'event already missing:{sid},{eid}')
            continue

        # Get id in the secondary redcap project
        try:
            rid2 = subj2id2[sid]
        except KeyError as err:
            logging.info(f'missing subject:{rid1}:{err}')
            missing_subjects.append(sid)
            continue

        # Get records from secondary redcap for this subject/event
        try:
            records2 = p2.export_records(records=[rid2], events=[eid])
        except Exception:
            import traceback
            traceback.print_exc()
            records2 = []

        # Find the best record to compare
        if len(records2) == 1:
            # Only one so use it
            r2 = records2[0]
            name2 = str(r2.get('redcap_repeat_instrument', ''))
            num2 = str(r2.get('redcap_repeat_instance', ''))
        else:
            # Find a matching date or instance number
            for e2 in records2:
                name2 = str(e2.get('redcap_repeat_instrument', ''))
                num2 = str(e2.get('redcap_repeat_instance', ''))

                if name1 != name2:
                    continue

                # First check specific identifiers
                if r1.get('vasf_timepoint', False):
                    if r1['vasf_timepoint'] == e2['vasf_timepoint']:
                        logging.info(f'vasf_timepoint match:{sid}:{eid}:{name1}:{num1}:{name2}:{num2}')
                        r2 = e2
                        break
                    else:
                        continue

                # Try to match a date
                for cur_field in date_fields:
                    # Does 1st have a value and does it match 2nd
                    if r1[cur_field] and r1[cur_field] == e2[cur_field]:
                        # This is the record
                        logging.info(f'date match:{sid}:{eid}:{name1}:{num1}:{cur_field}:{name2}:{num2}')
                        r2 = e2
                        break

                if r2:
                    break

            if r2 is None:
                for e2 in records2:
                    name2 = str(e2.get('redcap_repeat_instrument', ''))
                    num2 = str(e2.get('redcap_repeat_instance', ''))

                    if name1 == name2 and num1 == num2:
                        logging.info(f'{sid}:{eid}:instance match:{name1}:{num1}:{num2}')
                        r2 = e2
                        break

            if r2 is None:
                logging.info(f'NO MATCH:{sid}:{eid}:{name1}:{num1}')

        # Check for conflicts in best matching record 2
        if (name1 != name2) or (not name1 and (num1 != num2)):
            logging.info(f'skipping, name conflict:{sid},{eid},{rid1},{rid2},{name1},{num1},{name2},{num2}')
            r2 = None

        if r2:
            logging.info(f'compare_records:{sid},{eid},{rid1},{rid2},{name1},{num1},{name2},{num2}')
            (mism, misv, _m) = compare_records(r1, r2, compare_fields)
            mismatches += mism
            missing_values += misv
            match_count += _m
        else:
            _eid = eid

            if name1:
                _eid = f'{_eid}:{name1}'

            if num1:
                _eid = f'{_eid}:{num1}'

            logging.info(f'No record in double/second:{sid}:{_eid}')
            missing_events.append((sid, _eid))

    # Count results
    results['counts'] = {
        'missing_values': len(missing_values),
        'missing_events': len(missing_events),
        'missing_subjects': len(missing_subjects),
        'mismatches': len(mismatches),
        'matches': match_count,
    }

    # Convert subjects list of dicts, we do this here so we can keep
    # a simple list during the loop to check for already missing
    if missing_subjects:
        missing_subjects = [{'SUBJECT': s} for s in missing_subjects]
    else:
        missing_subjects = None

    # Convert events to list of dicts
    if missing_events:
        _keys = ['SUBJECT', 'EVENT']
        missing_events = [dict(zip(_keys, v)) for v in missing_events]
    else:
        missing_events = None

    # Append results
    results['missing_subjects'] = missing_subjects
    results['missing_events'] = missing_events
    results['missing_values'] = missing_values
    results['mismatches'] = mismatches

    # Append fields information
    results['fields'] = {}
    for k in ['compare', 'common', 'p1_only', 'p2_only', 'p2_nan']:
        if fields[k]:
            results['fields'][k] = [{'FIELD': v} for v in fields[k]]
        else:
            results['fields'][k] = [{'FIELDS': ''}]

    # Get project titles and ids
    p1_info = p1.export_project_info()
    p2_info = p2.export_project_info()
    name1 = '{} ({})'.format(p1_info['project_title'], p1_info['project_id'])
    name2 = '{} ({})'.format(p2_info['project_title'], p2_info['project_id'])

    # Build the info for pdf
    results['session'] = {
        'REDCap 1': name1,
        'REDCap 2': name2,
        'DATE': datetime.now().strftime("%Y-%m-%d"),
    }

    return results


def _simplify(full_string):
    simple_string = str(full_string)

    try:
        # Coerce date format
        simple_string = datetime.strptime(
            simple_string, '%m/%d/%y').strftime('%Y-%m-%d')
    except Exception:
        simple_string = str(full_string)
        simple_string = simple_string.replace(' ', '').replace('&', 'and')
        simple_string = simple_string[0:20].lower()

    return simple_string


def compare_records(r1, r2, fields, show_one_null=False, show_two_null=True):
    mismatches = []
    misvalues = []
    match_count = 0

    for k in sorted(fields):
        # Get value from the first redcap
        try:
            v1 = r1[k]
        except KeyError:
            logging.error(f'r1:KeyError:{k}')
            continue

        # Get the value from the second
        try:
            v2 = r2[k]
        except KeyError:
            logging.error(f'r2:KeyError:{k}')
            continue

        mis = {
            'SUBJECT': r1['sid'],
            'EVENT': r1['redcap_event_name'],
            'FIELD': k,
        }

        mis['REPEAT_INSTRUMENT'] = str(r1.get('redcap_repeat_instrument', ''))
        mis['1stREPEAT_INSTANCE'] = str(r1.get('redcap_repeat_instance', ''))
        mis['2ndREPEAT_INSTANCE'] = str(r2.get('redcap_repeat_instance', ''))

        if v1 == '':
            # First blank
            if show_one_null:
                mismatches.append(mis)
        elif v2 == '':
            # First has value, Second is blank
            if show_two_null:
                misvalues.append(mis)
        else:
            c1 = v1
            c2 = v2

            if k == 'country_birth':
                c1 = c1.replace('.', '').replace('nited ', '')[0:2].lower()
                c2 = c2.replace('.', '').replace('nited ', '')[0:2].lower()
            else:
                try:
                    if float(c1) == float(c2):
                        # Set them to be exactly the same
                        c2 = c1
                except Exception:
                    # Not a float, simplifiy string
                    c1 = _simplify(c1)
                    c2 = _simplify(c2)

            # Now compare the cleaned values
            if c1 == c2:
                match_count += 1
            else:
                # Both have values, but don't match, show the truncated values
                mis['1stVALUE'] = v1[0:50]
                mis['2ndVALUE'] = v2[0:50]
                logging.debug(f'mismatch:{mis}:{c1}:{c2}')
                mismatches.append(mis)

    return (mismatches, misvalues, match_count)


def write_results(results, pdf_file, excel_file):
    # Make the summary PDF
    make_pdf(results, pdf_file)

    # Save excel file with results
    write_excel(results, excel_file)


def test_finish(outdir, outpref):
    info = {}
    excel_file = os.path.join(outdir, f'{outpref}.xlsx')
    pdf_file = os.path.join(outdir, f'{outpref}.pdf')

    info['session'] = {
        'REDCap 1': '1',
        'REDCap 2': '2',
        'DATE': datetime.now().strftime("%Y-%m-%d"),
    }

    info['counts'] = {
        'fields_compare': 0,
        'missing_events': 0,
        'missing_subjects': 0,
        'mismatches': 0,
        'missing_values': 0,
    }

    info['missing_subjects'] = []
    info['missing_events'] = []
    info['mismatches'] = []
    info['missing_values'] = []

    info['fields'] = {
        'compare': [{'FIELD': v} for v in []],
        'common': [{'FIELD': v} for v in []],
        'p1_only': [{'FIELD': v} for v in []],
        'p2_only': [{'FIELD': v} for v in []],
        'p2_nan': [{'FIELD': v} for v in []],
    }

    # Write results
    write_results(info, pdf_file, excel_file)


def run_compare(p1, p2, pdf_file, excel_file):
    # Get the compare results
    results = compare_projects(p1, p2)

    # Write output files
    write_results(results, pdf_file, excel_file)



"""

Manager progress reports. Update will create any missing.

"""
from datetime import datetime
import tempfile
import logging

import pandas as pd

from .report import make_project_report


logger = logging.getLogger('garjus.progress')


def update(garjus, projects=None):
    """Update project progress."""
    for p in (projects or garjus.projects()):
        if p in projects:
            logger.debug(f'updating progress:{p}')
            update_project(garjus, p)


def update_project(garjus, project):
    """Update project progress."""
    progs = garjus.progress_reports(projects=[project])

    # what time is it? we will use this for naming
    now = datetime.now()

    # determine current month and year to get current monthly repot id
    cur_progress = now.strftime("%B%Y")

    # check that each project has report for current month with PDF and zip
    has_cur = any(d.get('progress_name') == cur_progress for d in progs)
    if not has_cur:
        logger.debug(f'making new progress record:{project}:{cur_progress}')
        make_progress(garjus, project, cur_progress, now)
    else:
        logger.debug(f'progress record exists:{project}:{cur_progress}')


def make_progress(garjus, project, cur_progress, now):
    with tempfile.TemporaryDirectory() as outdir:
        fnow = now.strftime("%Y-%m-%d_%H_%M_%S")
        pdf_file = f'{outdir}/{project}_report_{fnow}.pdf'
        zip_file = f'{outdir}/{project}_data_{fnow}.zip'

        make_project_report(garjus, project, pdf_file, zip_file, monthly=True)
        garjus.add_progress(project, cur_progress, now, pdf_file, zip_file)


def make_stats_csv(
    garjus, projects, proctypes, sesstypes, csvname, persubject=False, analysis=None
):
    """"Make the file."""
    df = pd.DataFrame()

    if not isinstance(projects, list):
        projects = projects.split(',')

    if proctypes is not None and not isinstance(proctypes, list):
        proctypes = proctypes.split(',')

    if sesstypes is not None and not isinstance(sesstypes, list):
        sesstypes = sesstypes.split(',')

    for p in sorted(projects):
        # Load stats
        stats = garjus.stats(
            p, proctypes=proctypes, sesstypes=sesstypes, persubject=persubject)
        df = pd.concat([df, stats])

    if analysis:
        # Get the list of subjects for specified analysis and apply as filter
        logger.info(f'analysis={analysis}')

        # Get the subject list from the analysis
        project, analysis_id = analysis.rsplit('_', 1)
        a = garjus.load_analysis(project, analysis_id)
        subjects = a['analysis_include'].splitlines()
        logger.debug(f'applying subject filter to include:{subjects}')
        df = df[df.SUBJECT.isin(subjects)]

        # Append rows for missing subjects and resort
        _subj = df.SUBJECT.unique()
        missing_subjects = [x for x in subjects if x not in _subj]
        if missing_subjects:
            logger.info(f'missing_subjects={missing_subjects}')
            df = pd.concat([
                df,
                pd.DataFrame(
                    missing_subjects,
                    columns=['SUBJECT']
                )
            ]).sort_values('SUBJECT')

    # Save file for this type
    logger.info(f'saving csv:{csvname}')
    df.to_csv(csvname, index=False)
